
/**
 * Override function. Get the message onLoadCompleted 'event'. A listener can be
 * used for msgLoaded and a multimessage url onLoad event, but there is no other
 * way to listen/observe a multimessage summarizeSelection completed event if a
 * multiselection is already loaded/selected.
 */
MessageDisplayWidget.prototype._onLoadCompleted = MessageDisplayWidget.prototype.onLoadCompleted;
MessageDisplayWidget.prototype.onLoadCompleted = function() {
  this._onLoadCompleted();

TotalQuickFilter.log.debug("MessageDisplayWidget.onLoadCompleted:  ");
  let h = document.getElementById("tqf-terms");
  if (h && h.checked)
    HiLighter.doHilight("onloadcompleted", null, HiLighter.highlightBrowserElement);
};

nsContextMenu.prototype._initItems = nsContextMenu.prototype.initItems;
nsContextMenu.prototype.initItems = function(aPopup) {

  // Only display context items useful for selected text.
  if (this.isContentSelected)
    this.hideMailItems = true;

  this._initItems();
};

/**
 * Override function. Get rid of odd code which prevents ORing grouped terms.
 */
SearchSpec.prototype._groupifyTerms = function(aTerms, aCloneTerms) {
  return aTerms;
};

/**
 * Lot of junk code here, to go along with the junky design.
 */
QuickFilterBarMuxer.onOverFlow = function() {};
QuickFilterBarMuxer.onWindowResize = function() {};

/**
 * If we've been disabled or removed, reset the quickfilter session. Do this
 * here rather than a quit observer.
 */
QuickFilterBarMuxer.onTabPersist = function(aTab) {
TotalQuickFilter.log.debug("onTabPersist: reset session - "+
                           TotalQuickFilter.AddonListener.reset);
  let filterer = ("quickFilter" in aTab._ext) ? aTab._ext.quickFilter : null;
  if (filterer && !TotalQuickFilter.AddonListener.reset)
    return filterer.persistToObj();
  return null;
};

/**
 * Keep a searching state to avoid running dom state code unnecessarily, and
 * fix to get rid of permanent Searching in statusbar..
 */
QuickFilterBarMuxer.isSearching = null;
QuickFilterBarMuxer.onSearching = function(aFolderDisplay, aIsSearching) {
TotalQuickFilter.log.debug("onSearching: aIsSearching - "+aIsSearching);
//  this.isSearching = null;
  if (!aFolderDisplay.active)
    return;

  this.isSearching = aIsSearching;

  if (!aIsSearching) {
    let t = document.getElementById("qfb-filter-bar-textbox");
    (t && t.removeAttribute("searching"))
    statusFeedback.showStatusString("");
  }
  else {
    // - Update match status.
    this.reflectFiltererResults(this.activeFilterer, aFolderDisplay);
  }
};

/**
 * Patch to enable inverted filtering and prevent searches on textfilter button
 * action when there's no text for the button.
 */
QuickFilterBarMuxer._bindUI = function() {
  for (let [, filterDef] in Iterator(QuickFilterManager.filterDefs)) {
    let domNode = document.getElementById(filterDef.domId);
    if (!domNode)
      continue;
    // the loop let binding does not latch, at least in 1.9.2
    let latchedFilterDef = filterDef;

    let handler;
    if (!("onCommand" in filterDef)) {
      handler = function(aEvent) {
        try {
          let postValue = domNode.checked ?
                          domNode.hasAttribute("inverted") ?
                          false : true : null;
          QuickFilterBarMuxer.activeFilterer.setFilterValue(
            latchedFilterDef.name, postValue);
          QuickFilterBarMuxer.deferredUpdateSearch();
        }
        catch (ex) {
          logException(ex);
        }
      };
    }
    else {
      handler = function(aEvent) {
        let filterValues = QuickFilterBarMuxer.activeFilterer.filterValues;
        let preValue = (latchedFilterDef.name in filterValues) ?
                         filterValues[latchedFilterDef.name] : null;
        let [postValue, update] =
          latchedFilterDef.onCommand(preValue, domNode, aEvent, document);
        QuickFilterBarMuxer.activeFilterer.setFilterValue(
          latchedFilterDef.name, postValue, !update);
        if (update)
          QuickFilterBarMuxer.deferredUpdateSearch();
      };
    }
    domNode.addEventListener("command", handler, false);

    if ("domBindExtra" in filterDef)
      filterDef.domBindExtra(document, this, domNode);
  }
};

/**
 * Patch for correct button state for inverted filtering.
 */
QuickFilterBarMuxer.reflectFiltererState = function(aFilterer, aFolderDisplay, aFilterName) {
TotalQuickFilter.log.debug("reflectFiltererState: START visible - "+aFilterer.visible);
  // If we aren't visible then there is no need to update the widgets.
  if (aFilterer.visible) {
    let filterValues = aFilterer.filterValues;
    for (let [, filterDef] in Iterator(QuickFilterManager.filterDefs)) {
      // If we only need to update one state, check and skip as appropriate.
      if (aFilterName && filterDef.name != aFilterName)
        continue;

      let domNode = document.getElementById(filterDef.domId);
      if (!domNode)
        continue;
      let value = (filterDef.name in filterValues) ? filterValues[filterDef.name] :
                                                     null;
TotalQuickFilter.log.debug("reflectFiltererState: id:value:inverted - "+
                           domNode.id+":"+
                           (value == null ? value : value.toSource()) +":"+
                           domNode.hasAttribute("inverted"));
      if (!("reflectInDOM" in filterDef)) {
        domNode.checked = value !== null;
        if (value === false)
          domNode.setAttribute("inverted", true);
        else
          domNode.removeAttribute("inverted");
      }
      else
        filterDef.reflectInDOM(domNode, value, document, this);
    }
  }

  this.reflectFiltererResults(aFilterer, aFolderDisplay);

  document.getElementById("quick-filter-bar").collapsed = !aFilterer.visible;
  document.getElementById("view_toolbars_popup_quickFilterBar")
          .setAttribute("checked", aFilterer.visible);

  let infoButton = document.getElementById("tqf-quickFilterInfo-button");
  if (!aFilterer.visible && infoButton.checked) {
    infoButton.checked = false;
    TotalQuickFilter.QuickFilterInfo(infoButton);
  }
};

/**
 * Override function.  For UI responsiveness, the current code is ineffective.
 * Make updateSearch() on a timeout, and note 10ms is not enough where the
 * attempt was made in deferredUpdateSearch().
*/
QuickFilterBarMuxer._updateSearch = QuickFilterBarMuxer.updateSearch;
QuickFilterBarMuxer.updateSearch = function(aTab) {
  let tab = aTab || this.tabmail.currentTabInfo;
  // Bail if things don't really exist yet.
  if (!tab.folderDisplay || !tab.folderDisplay.view.search)
    return;
TotalQuickFilter.log.debug("QuickFilterBarMuxer.updateSearch: tabTitle:busy:name - "+
                           tab.title+":"+tab.busy+":"+tab.mode.name);

  let b = document.getElementById("quick-filter-bar");
  (tab.folderDisplay.active && b && b.setAttribute("filterActive", "searching"))

  setTimeout(function() {
    QuickFilterBarMuxer._updateSearch(aTab);
    if (!tab.folderDisplay.view.search.userTerms)
      (b && b.removeAttribute("filterActive"))
TotalQuickFilter.log.debug("QuickFilterBarMuxer.updateSearch:\n"+
                           tab.folderDisplay.view.search.prettyString());
  }, 20);
};

QuickFilterManager._clearAllFilterValues = QuickFilterManager.clearAllFilterValues;
QuickFilterManager.clearAllFilterValues = function(aFilterValues) {
  QuickFilterManager._clearAllFilterValues(aFilterValues);

  aFilterValues.text.curTerms = null;
  TotalQuickFilter.setHilightButton("textfiltercommand");
};


/*******************************************************************************
 * Tags functions. The TagFacetingFilter symbol is not exported. Support AND/OR
 * tag selection with the patch from Bug 734221, which is included/modified.
 * Also support inverted tag filtering.
 ******************************************************************************/
/**
 * Override function. Support AND/OR tag selection.
 */
// TagFacetingFilter.isSimple =
QuickFilterManager.filterDefsByName.tags.isSimple = function(aFilterValue) {
  // it's the simple case if the value is just a boolean
  if (typeof(aFilterValue) != "object")
    return true;
  // but also if the object contains no non-null values
  let simpleCase = true;
  for (let [key, value] in Iterator(aFilterValue.tags)) {
    if (value !== null) {
      simpleCase = false;
      break;
    }
  }

  return simpleCase;
};

/**
 * Override function. Support AND/OR tag selection.
 */
// TagFacetingFilter.appendTerms =
QuickFilterManager.filterDefsByName.tags.appendTerms = function(aTermCreator, aTerms, aFilterValue) {
TotalQuickFilter.log.debug("TagFacetingFilter.appendTerms: aFilterValue - "+
                           aFilterValue.toSource());
  if (aFilterValue == null)
    return null;

  let term, value;

  // just the true/false case
  if (this.isSimple(aFilterValue)) {
    term = aTermCreator.createTerm();
    term.attrib = Components.interfaces.nsMsgSearchAttrib.Keywords;
    value = term.value;
    value.str = "";
    term.value = value;
    term.op = aFilterValue ?
                Components.interfaces.nsMsgSearchOp.IsntEmpty :
                Components.interfaces.nsMsgSearchOp.IsEmpty;
    term.booleanAnd = true;
    aTerms.push(term);

    // we need to perform faceting if the value is literally true.
    // NOTE: Fix Bug 612800; always return |this| listener if tags are active,
    // to reflect tag buttons per filtered messages results, sort of obvious.
//    if (aFilterValue === true)
      return this;
  }
  else {
    let firstIncludeClause = true, firstExcludeClause = true;
    let lastIncludeTerm = null;
    term = null;

    let excludeTerms = [];

    let mode = aFilterValue.mode;
TotalQuickFilter.log.debug("TagFacetingFilter.appendTerms: mode - "+mode);
    for (let [key, shouldFilter] in Iterator(aFilterValue.tags)) {
      if (shouldFilter !== null) {
        term = aTermCreator.createTerm();
        term.attrib = Components.interfaces.nsMsgSearchAttrib.Keywords;
        value = term.value;
        value.attrib = term.attrib;
        value.str = key;
        term.value = value;
        if (shouldFilter) {
          term.op = nsMsgSearchOp.Contains;
          // AND for the group. Inside the group we also want AND if the
          // mode is set to "All of".
          term.booleanAnd = firstIncludeClause || (mode === "AND");
          term.beginsGrouping = firstIncludeClause;
          aTerms.push(term);
          firstIncludeClause = false;
          lastIncludeTerm = term;
        }
        else {
          term.op = nsMsgSearchOp.DoesntContain;
          // you need to not include all of the tags marked excluded.
          term.booleanAnd = true;
          term.beginsGrouping = firstExcludeClause;
          excludeTerms.push(term);
          firstExcludeClause = false;
        }
      }
    }
    if (lastIncludeTerm)
      lastIncludeTerm.endsGrouping = true;

    // if we have any exclude terms:
    // - we might need to add a "has a tag" clause if there were no explicit
    //   inclusions.
    // - extend the exclusions list in.
    if (excludeTerms.length) {
      // (we need to add has a tag)
      if (!lastIncludeTerm) {
        term = aTermCreator.createTerm();
        term.attrib = Components.interfaces.nsMsgSearchAttrib.Keywords;
        value = term.value;
        value.str = "";
        term.value = value;
        term.op = Components.interfaces.nsMsgSearchOp.IsntEmpty;
        term.booleanAnd = true;
        aTerms.push(term);
      }

      // (extend in the exclusions)
      excludeTerms[excludeTerms.length-1].endsGrouping = true;
      aTerms.push.apply(aTerms, excludeTerms);
    }
  }
  for (let nsiTerm of aTerms) {
    TotalQuickFilter.log.debug("TagFacetingFilter.appendTerms: END aTerms," +
                               " beginGrp:termStr:bool:endGrp - "+
                               nsiTerm.beginsGrouping + ":" +
                               nsiTerm.termAsString + ":" +
                               (nsiTerm.booleanAnd ? "AND" : "OR") + ":" +
                               nsiTerm.endsGrouping);
  }
  return null;
};

/**
 * Override function.  Support AND/OR tag selection; maintain negated state.
 */
// TagFacetingFilter.onSearchDone =
QuickFilterManager.filterDefsByName.tags.onSearchDone = function(aCurState, aKeywordMap, aStatus) {
  // we are an async operation; return tags off or negated state.
  if (aCurState == null || aCurState == false)
    return [aCurState, false, false];

  // only propagate things that are actually tags though!
  let outKeyMap = {tags: {}};
  let tags = MailServices.tags.getAllTags({});
  for (let tag of tags) {
    if (tag.key in aKeywordMap)
      outKeyMap.tags[tag.key] = aKeywordMap[tag.key];
  }

  return [outKeyMap, true, false];
};

/**
 * Override function. Support AND/OR tag selection.
 */
// TagFacetingFilter.domBindExtra =
QuickFilterManager.filterDefsByName.tags.domBindExtra = function(aDocument, aMuxer, aNode) {
  // Tag filtering mode button (All of/Any of)
  function commandHandler(aEvent) {
    // Toggle.
    let newVal = aEvent.target.value == "AND" ? false : true;
    TotalQuickFilter.setTagBooleanButton(newVal);
    let b = document.getElementById("quick-filter-bar-tab-bar");
    let [, firstCheckedChildId,,] = TotalQuickFilter.getCheckedChildren(null, b);

    let filterValue = aMuxer.getFilterValueForMutation("tags");
    filterValue.mode = aEvent.target.value;
    if (firstCheckedChildId)
      aMuxer.updateSearch();
  }
  aDocument.getElementById("tqf-boolean-mode")
           .addEventListener("command", commandHandler);
},

/**
 * Override function. Support AND/OR tag selection and inverted filter.
 */
// TagFacetingFilter._populateTagBar =
QuickFilterManager.filterDefsByName.tags._populateTagBar = function(aState, aDocument, aMuxer) {
  let tagbar = aDocument.getElementById("quick-filter-bar-tab-bar");
  let keywordMap = aState.tags;

  if (!tagbar)
    return;

  let qbm = aDocument.getElementById("tqf-boolean-mode");
  if (qbm) {
    if (aState.mode)
      qbm.value = aState.mode;
    else
      aState.mode = qbm.value;
  }

  function commandHandler(aEvent) {
    let tagKey = aEvent.target.getAttribute("value");
    let state = aMuxer.getFilterValueForMutation("tags");
    state.tags[tagKey] = aEvent.target.checked ? true : null;
    aEvent.target.removeAttribute("inverted");
    aMuxer.updateSearch();
  };

  function rightClickHandler(aEvent) {
    // we need to toggle the checked state ourselves
    aEvent.target.checked = !aEvent.target.checked;

    let tagKey = aEvent.target.getAttribute("value");
    let state = aMuxer.getFilterValueForMutation("tags");
    state.tags[tagKey] = aEvent.target.checked ? false : null;
    if (aEvent.target.checked)
      aEvent.target.setAttribute("inverted", "true");
    else
      aEvent.target.removeAttribute("inverted");
    aMuxer.updateSearch();
    aEvent.stopPropagation();
    aEvent.preventDefault();
  }

  // -- nuke existing exposed tags, but not the mode selector (which is first)
  while (tagbar.lastChild && tagbar.lastChild.tagName == "toolbarbutton")
    tagbar.removeChild(tagbar.lastChild);

  let addCount = 0;

  // -- create an element for each tag
  let tags = MailServices.tags.getAllTags({});
  for (let tag of tags) {
    if (tag.key in keywordMap) {
      addCount++;

      // Keep in mind that the XBL does not get built for dynamically created
      // elements such as these until they get displayed, which definitely
      // means not before we append it into the tree.
      let button = aDocument.createElement("toolbarbutton");

      button.setAttribute("id", "qfb-tag-" + tag.key);
      button.addEventListener("command", commandHandler, false);
      button.addEventListener("click",
                              function(event) {
                                TotalQuickFilter.clickFunc(event, this,
                                                           commandHandler,
                                                           rightClickHandler);
                              });
      button.addEventListener("keypress",
                              function(event) {
                                TotalQuickFilter.keyFunc(event, this,
                                                         commandHandler,
                                                         rightClickHandler);
                              });
      button.addEventListener("focus", TotalQuickFilter.focusHandler);
      button.addEventListener("mouseenter", TotalQuickFilter.focusHandler);

      button.setAttribute("type", "checkbox");
      if (keywordMap[tag.key] !== null) {
        button.setAttribute("checked", "true");
        if (!keywordMap[tag.key])
          button.setAttribute("inverted", "true");
      }
      button.setAttribute("label", tag.tag);
      button.setAttribute("value", tag.key);
      // everybody always gets to be an qfb-tag-button.
      button.setAttribute("class", "toolbarbutton-1 qfb-tag-button");
      let color = tag.color;
      if (color)
        button.classList.add("lc-" + color.substr(1));
      button.setAttribute("context", "");
      tagbar.appendChild(button);
    }
  }

  tagbar.collapsed = !addCount;

  if (addCount)
    tagbar.setAttribute("haschild", true);
  else
    tagbar.removeAttribute("haschild");
TotalQuickFilter.log.debug("QuickFilterManager.tags._populateTagBar: addCount - "+addCount);
};

/**
 * Override function. Support inverted filter.
 */
// TagFacetingFilter.onCommand =
QuickFilterManager.filterDefsByName.tags.onCommand = function(aState, aNode, aEvent, aDocument) {
TotalQuickFilter.log.debug("QuickFilterManager.tags.onCommand: checked - "+
                           aNode.checked);
  let checked = aNode.checked ? aNode.hasAttribute("inverted") ? false : true : null;
  if (!checked)
    aDocument.getElementById("quick-filter-bar-tab-bar").collapsed = true;

  // return ourselves if we just got checked to have
  //  onSearchStart/onSearchMessage/onSearchDone get to do their thing.
  return [checked, true];
};

/**
 * Override function. Support inverted filter.
 */
// TagFacetingFilter.reflectInDOM =
QuickFilterManager.filterDefsByName.tags.reflectInDOM = function(aNode, aFilterValue, aDocument, aMuxer) {
TotalQuickFilter.log.debug("QuickFilterManager.tags.reflectInDOM: aNode - "+ aNode.id);
  aNode.checked = aFilterValue !== null ? true : false;

  if ((aFilterValue != null) && (typeof(aFilterValue) == "object"))
    this._populateTagBar(aFilterValue, aDocument, aMuxer);
  else
    aDocument.getElementById("quick-filter-bar-tab-bar").collapsed = true;

  aDocument.getElementById("quick-filter-bar-collapsible-buttons")
           .removeAttribute("shrink");
};


/*******************************************************************************
 * Textfilter functions.
 ******************************************************************************/
/**
 * Override function. Remove upsell code; don't run command unnecessarily on
 * textfilter button action.
 */
MessageTextFilter.domBindExtra = function(aDocument, aMuxer, aNode) {
  // -- platform-dependent placeholder setup
  aNode.setAttribute(
    "placeholder",
    aNode.getAttribute("emptytextbase")
         .replace("#1", aNode.getAttribute(Application.platformIsMac ?
                                           "keyLabelMac" : "keyLabelNonMac")));
  // force an update of the emptytext now that we've updated it.
  aNode.value = "";

  // Textfilter buttons.
  function commandHandler(aEvent) {
    let state = aMuxer.getFilterValueForMutation(MessageTextFilter.name);
    let filterDef = MessageTextFilter.textFilterDefsByDomId[aEvent.target.id];
    state.states[filterDef.name] = aEvent.target.checked;

    // Check if search should run.
    if (TotalQuickFilter.shouldUpdate(aEvent.target))
      aMuxer.updateSearch();
  }

  for (let [, textFilter] in Iterator(this.textFilterDefs)) {
    aDocument.getElementById(textFilter.domId)
             .addEventListener("command", commandHandler, false);
  }
};

/**
 * Override function. Mainly, don't waste searches on failed edits or spaces
 * and return false in such cases.
 */
MessageTextFilter._onCommand = MessageTextFilter.onCommand;
MessageTextFilter.onCommand = function(aState, aNode, aEvent, aDocument) {
TotalQuickFilter.log.debug("MessageTextFilter.onCommand: aState - "+
                           aState.toSource());
  let terms = TotalQuickFilter.onFilter(aNode.value, aState);
  if (terms === false ||
      (aState["curTerms"] && terms.length == aState["curTerms"].length &&
       terms.toString() == aState["curTerms"].toString()) && aNode.value)
    return [aState, false];

  if (!aNode.value) {
    aState.text = aNode.value;
    aState["curTerms"] = null;
  }
  else {
    aState["curTerms"] = terms;
  }

  aNode.setAttribute("value", aNode.value);

  let p = document.getElementById("quick-filter-bar-textfilter-bar");
  let h = document.getElementById("tqf-terms");
  let [, firstCheckedChildId,,] = TotalQuickFilter.getCheckedChildren(null, p);
  if (!firstCheckedChildId && (!h || (h && !h.checked))) {
    if (aNode.value)
      aState.text = aNode.value;
    return [aState, false];
  }

  if (h && h.checked)
    TotalQuickFilter.setHilightButton("textfiltercommand");
  else
    TotalQuickFilter.setTermsButtons();

  aNode.focus();

  return this._onCommand(aState, aNode, aEvent, aDocument);
};

/**
 * Override function.
 */
MessageTextFilter._reflectInDOM = MessageTextFilter.reflectInDOM;
MessageTextFilter.reflectInDOM = function(aNode, aFilterValue, aDocument, aMuxer, aFromPFP) {
  if (aMuxer.isSearching)
    return;
TotalQuickFilter.log.debug("MessageTextFilter.reflectInDOM: nodeVal:textVal - "+
                           aNode.value+":"+aFilterValue.text);

  aNode.removeAttribute("invalid");
  aNode.setAttribute("value", aFilterValue.text || "");
  aNode.setAttribute("timeout", aNode.getAttribute("defaulttimeout"));

  this._reflectInDOM(aNode, aFilterValue, aDocument, aMuxer, aFromPFP);
};

/**
 * Override function. Support OR and Negation searches within a textfilter.
 */
MessageTextFilter.appendTerms = function(aTermCreator, aTerms, aFilterValue) {
TotalQuickFilter.log.debug("MessageTextFilter.appendTerms: " +
                           "text:curTerms:value:states - "+
                           aFilterValue.text+":"+
                           aFilterValue.curTerms+":"+
                           document.getElementById("qfb-qs-textbox").value+":"+
                           aFilterValue.states.toSource());
  let term, value;

  if (aFilterValue.text) {
    TotalQuickFilter.getCurTerms(aFilterValue);
    let mode = aFilterValue.mode;
    let firstFilter = true;
TotalQuickFilter.log.debug("MessageTextFilter.appendTerms: domIds - " +
                           TotalQuickFilter.orderedTextFilterDomIds.toSource());
    let i = 0;
    for (let [tfDomId, tfName] in Iterator(TotalQuickFilter.orderedTextFilterDomIds)) {
TotalQuickFilter.log.debug("MessageTextFilter.appendTerms: tfName:state - "+
                           tfName+":"+aFilterValue.states[tfName]);
      let tfButton = document.getElementById(tfDomId);
      if (!aFilterValue.states[tfName] || (tfButton && tfButton.disabled))
        // Button not active.
        continue;

      let tfDef = this.textFilterDefs[tfName];
      let phrases = aFilterValue.curTerms[i++] || [];
TotalQuickFilter.log.debug("MessageTextFilter.appendTerms: CONTINUE tfName:domId - "+
                           tfName+":"+tfDomId);
      if (!phrases.length || !tfDef)
        continue;

      let nextFilter = true;
      let orTermNext = false;
      let negated = false;
      let or = false;
      let quoted = false;
      let mode = (aFilterValue.modes && (tfName in aFilterValue.modes) &&
                  aFilterValue.modes[tfName] === false) ? false : true;
TotalQuickFilter.log.debug("MessageTextFilter.appendTerms " +
                           "firstFilter:nextFilter:orTermNext - "+
                           firstFilter+":"+nextFilter+":"+orTermNext);
TotalQuickFilter.log.debug("MessageTextFilter.appendTerms mode:modes - "+
                           (("modes" in aFilterValue) ?
                            (tfName in aFilterValue.modes)  ?
                             (aFilterValue["modes"][tfName]+":"+
                              aFilterValue["modes"].toSource()) : null : null));
      for (let [, phrase] in Iterator(phrases)) {
//TotalQuickFilter.log.debug("MessageTextFilter.appendTerms phrase - "+ phrase);
        [negated, or, quoted, phrase] = TotalQuickFilter.parseTerm(phrase);
//TotalQuickFilter.log.debug("MessageTextFilter.appendTerms neg:or:phrase - "+
//                           negated+":"+orTermNext+":"+phrase);
        if (or) {
          orTermNext = true;
          continue;
        }

        term = aTermCreator.createTerm();
        term.attrib = tfDef.attrib;
        value = term.value;
        value.attrib = tfDef.attrib;
        value.str = phrase;
        term.value = value;
        term.op = negated ? nsMsgSearchOp.DoesntContain : nsMsgSearchOp.Contains;
        // OR/AND among term groups based on mode.  AND within term's phrases,
        // unless explicit OR.
        term.booleanAnd = firstFilter ? true :
                          nextFilter ? mode :
                          orTermNext ? false :
                          negated ? true :
                          true;
        term.beginsGrouping = firstFilter || nextFilter;
        aTerms.push(term);
        firstFilter = false;
        nextFilter = false;
        orTermNext = false;
      }
      if (term)
        term.endsGrouping = true;
    }
  }

  for (let nsiTerm of aTerms) {
    TotalQuickFilter.log.debug("MessageTextFilter.appendTerms: END aTerms, " +
                               "beginGrp:termStr:bool:endGrp - "+
                               nsiTerm.beginsGrouping + ":" +
                               nsiTerm.termAsString + ":" +
                               (nsiTerm.booleanAnd ? "AND" : "OR") + ":" +
                               nsiTerm.endsGrouping);
  }
};

/**
 * Override function. Get rid of upsell, gloda search won't understand the
 * OR and NOT syntax. Plus, it's annoying.
 */
MessageTextFilter.postFilterProcess = function(aState, aViewWrapper, aFiltering) {
  return [aState, "nosale", false];
};


/*******************************************************************************
 * Add support for more filter options.
 ******************************************************************************/
MessageTextFilter.defineTextFilter({
  name: "to",
  domId: "tqf-qs-to",
  attrib: nsMsgSearchAttrib.To,
  defaultState: false,
});

MessageTextFilter.defineTextFilter({
  name: "cc",
  domId: "tqf-qs-cc",
  attrib: nsMsgSearchAttrib.CC,
  defaultState: false,
});

MessageTextFilter.defineTextFilter({
  name: "bcc",
  domId: "tqf-qs-bcc",
  attrib: nsMsgSearchAttrib.Custom,
  defaultState: false,
});

MessageTextFilter.defineTextFilter({
  name: "alladdresses",
  domId: "tqf-qs-alladdresses",
  attrib: nsMsgSearchAttrib.AllAddresses,
  defaultState: false,
});

//MessageTextFilter.defineTextFilter({
//  name: "date",
//  domId: "tqf-qs-date",
//  attrib: nsMsgSearchAttrib.Date,
//  defaultState: false,
//});

//MessageTextFilter.defineTextFilter({
//  name: "received",
//  domId: "tqf-qs-received",
//  attrib: nsMsgSearchAttrib.OtherHeader,
//  defaultState: false,
//});

//MessageTextFilter.defineTextFilter({
//  name: "hilighter",
//  domId: "tqf-terms",
//  defaultState: false,
//});


/*******************************************************************************
 * Hilighter functions.
 ******************************************************************************/
QuickFilterManager.defineFilter({
  name: "hilighter",
  domId: "tqf-terms",
  defaultState: false,
  appendTerms: function(aTermCreator, aTerms, aFilterValue) {
    // Invoked when not wanted (e.g. new tab, with no way to know without even
    // more patching), so tagalong textfilter oncommand for text input changes.
  },
  onCommand: function(aState, aNode, aEvent, aDocument) {
    let checked = aNode.checked ? true : null;
TotalQuickFilter.log.debug("QuickFilterManager.hilighter.onCommand: " +
                           "aState:checked - "+ aState+":"+checked);
    QuickFilterBarMuxer.activeFilterer.setFilterValue(this.name, checked, true);
    TotalQuickFilter.setHilightButton("command");
    return [checked, false];
  },
  reflectInDOM: function(aNode, aFilterValue, aDocument, aMuxer) {
TotalQuickFilter.log.debug("QuickFilterManager.hilighter.reflectInDOM: "+
                           aFilterValue);
    aNode.checked = aFilterValue === true;
  },
  clearState: function(aState) {
TotalQuickFilter.log.debug("QuickFilterManager.hilighter.aState: "+aState);
    return [aState, false];
  },
});


/**
 * Support SearchForSender extension.
 */
if ("SearchForSender" in window) {
  SearchForSender.search = function(searchTerm, useSender, useSubject) {
    let tfName = useSender ? "sender" : useSubject ? "subject" : null;
    if (useSubject)
      searchTerm = '"' + searchTerm.replace(/("|,)/g, "\\$1") + '"';

    TotalQuickFilter.searchFor(searchTerm, tfName);
  };

}

window.addEventListener("load", TotalQuickFilter.onLoad, true);
